from easy_spider.network.request import Request
from easy_spider.network.client import AsyncClient, SimpleClient
from easy_spider.network.response import Response, HTMLResponse
from easy_spider.extractors.extractor import SimpleBSExtractor
from easy_spider import tool
from abc import ABC, abstractmethod
from easy_spider.filters.build_in import html_filter, all_pass_filter
from typing import Iterable


class Spider(ABC):

    def __init__(self):
        self._start_targets = []
        self.num_threads = 1
        self._filter = html_filter
        self.extractor = SimpleBSExtractor()

    @property
    def start_targets(self):
        return self._start_targets

    @start_targets.setter
    def start_targets(self, targets):
        self._start_targets = self.from_url_iter(targets, use_default_params=False)

    @property
    def filter(self): return self._filter

    @filter.setter
    def filter(self, filter):
        self._filter = filter or all_pass_filter

    def _set_default_request_param(self, request):
        """
            若 self 中存在与 request 相同名称的属性，则将其值复制给 request
        """
        for attr in tool.get_public_attr(request):
            hasattr(self, attr) and tool.copy_attr(attr, self, request)
        return request

    def from_url(self, url: str, use_default_params=True):
        """
            先根据 request_like 对象生成 request 对象，
            若 use_default_params 为 True，则使用默认值覆盖 request 中的值，
            否则直接方法 request 对象
        """
        request = Request.of(url)
        if use_default_params:
            request = self._set_default_request_param(request)
        return request

    def from_url_iter(self, urls: Iterable[str], use_default_params=True):
        yield from (self.from_url(url, use_default_params) for url in urls)

    @staticmethod
    def _nothing():
        return range(0)

    @abstractmethod
    def crawl(self, request: Request):
        pass


# class MultiThreadSpider(Spider, SimpleClient):
#
#     def __init__(self, handlers):
#         super().__init__(handlers)
#         self.start_requests = []
#
#     def crawl(self, request: Request):
#         response = self.do_request(request)  # 发送请求
#         return request.handler(response)


class AsyncSpider(Spider, AsyncClient):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def handle(self, response: Response):
        """
        从 Response 中提取 Request，提取的 Request 将会设置默认参数
        """
        if isinstance(response, HTMLResponse):
            yield from self.from_url_iter(self.extractor.extract(response))
        else:
            yield from self._nothing()

    async def crawl(self, request: Request):
        """
        爬虫主要逻辑
        """
        response = await self.do_request(request)
        request.handler = request.handler if request.handler else self.handle
        new_requests = request.handler(response)
        if not new_requests:
            return self._nothing()
        return filter(lambda r: self.filter.accept(r.uri),
                      (Request.of(new_request) for new_request in new_requests))



import unittest
from easy_spider.network.request import Request
from easy_spider.core.env import async_env
from easy_spider.core.spider import AsyncSpider
from test import mock_env


class MySpider(AsyncSpider):

    def __init__(self):
        super().__init__()
        self.num_threads = 4

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class TestCore(unittest.TestCase):

    def setUp(self) -> None:
        self.my_spider = MySpider()
        self.my_spider.start_targets = ["http://localhost:5000/test_extract"]

    def test_set_default_method(self):
        r = Request("test")
        self.my_spider.method = 'POST'
        self.my_spider._set_default_request_param(r)
        self.assertEqual(r.method, 'POST')

    def test_core(self):
        self.my_spider.method = 'GET'
        async_env.run(self.my_spider)


if __name__ == '__main__':
    unittest.main()



from abc import ABC, abstractmethod
from easy_spider.network.request import Request
from easy_spider.network.response import (Response, TextResponse, HTMLResponse)
from requests import Session
from aiohttp import ClientSession, ClientTimeout


class Client(ABC):

    @abstractmethod
    def do_request(self, resource: Request) -> Response: pass

    @staticmethod
    def to_response(content, url, headers):
        content_type = headers.get("Content-Type")
        args = (content, url, headers)
        if content_type:
            if "text/html" in content_type:
                return HTMLResponse(*args)
            elif "text/" in content_type:
                return TextResponse(*args)
        return Response(*args)


class SimpleClient(Client):
    def __init__(self, session):
        self._session: Session = session

    def do_request(self, resource: Request) -> Response:
        raw_response = self._session.get(resource.uri)
        return self.to_response(raw_response.content, raw_response.url, raw_response.headers)


class AsyncClient(Client):

    def __init__(self, session=None):
        self._session: ClientSession = session

    def set_session(self, session):
        self._session = session

    async def do_request(self, request: Request):
        assert request
        common_params = dict(
            headers=request.headers,
            cookies=request.cookies,
            timeout=ClientTimeout(request.timeout),
            params=request.params
        )
        data_format = request.data_format.lower()
        if data_format == "form":
            common_params["data"] = request.data
        elif data_format == "json":
            common_params["json"] = request.data_format
        else:
            raise ValueError("未知数据类型 {} ，仅支持 json 或 form")
        raw_response = await self._session.request(request.method, request.uri, **common_params)
        content = await raw_response.content.read()
        return self.to_response(content, str(raw_response.url), raw_response.headers)



import unittest
from easy_spider.core.spider import AsyncSpider
from easy_spider.network.request import Request
from test.mock_env import run_and_get_result, env


class MySpider(AsyncSpider):

    def __init__(self):
        super().__init__()
        self.num_threads = 4

    def handle(self, response):
        print(response.bs.title)
        yield from super().handle(response)


class TestSpider(unittest.TestCase):

    async def async_spider(self):
        r = Request("http://localhost:5000/test_extract")
        spider = MySpider()
        spider.set_session(env.session)
        requests = await spider.crawl(r)
        for request in requests:
            print(request)

    def test_async_spider(self):
        run_and_get_result(self.async_spider())


if __name__ == '__main__':
    unittest.main()



from abc import ABC
from urllib.parse import urljoin
from w3lib.encoding import html_to_unicode, http_content_type_encoding
from bs4 import BeautifulSoup
from cached_property import cached_property


class Response(ABC):
    def __init__(self, body, url, headers):
        self.body = body
        self.url = url
        self.headers = headers

    def url_join(self, other_url):
        return urljoin(self.url, other_url)


class TextResponse(Response):
    """
        将 requests.response 对象解码为 TextResponse
        首先从 headers["Content-Type"] 中推断编码类型
        若不存在，则使用 w3lib.encoding.http_content_type_encoding 解码并推断编码类型
        w3lib.encoding.http_content_type_encoding 推断编码类型时需要耗费额外计算量
        若已知确定的编码格式，可以直接指定。
        如:
            r.encoding = 'gb2312'
            r.text # 将采用 gb2312 进行解码
        在某些情况，一些网站会给出与网站内容不符合的编码，这时候会导致乱码。此时可以将 r.encoding 设置为 None
        使用自动从网页内容推断编码
            r.encoding = None
            r.text # 自动从网页内容推断编码，乱码概率小，但会引入额外计算量
    """
    def __init__(self, body, url, headers):
        super().__init__(body, url, headers)
        self._encoding = self._infer_encoding_from_content_type()  # 从 Content-Type 中推断编码
        self._text = None

    def _decoding(self):
        charset = f'charset={self._encoding}'
        enc, text = html_to_unicode(charset, self.body)
        self._encoding = enc
        self._text = text

    def _infer_encoding_from_content_type(self):
        content_type = self.headers.get("Content-Type")
        if content_type:
            return http_content_type_encoding(content_type)
        return None

    @property
    def text(self):
        if self._text is None:
            self._decoding()
        return self._text

    @property
    def encoding(self):
        # 如果 Content-Type 从无法推测编码
        # 则使用 w3lib.encoding.http_content_type_encoding
        # 使用详情见: https://w3lib.readthedocs.io/en/latest/w3lib.html
        if self._encoding is None:
            self._decoding()
        return self._encoding

    @encoding.setter
    def encoding(self, encoding):
        self._encoding = encoding
        self._text = None  # 当用户重设编码时，需要清除 text 属性缓存


class HTMLResponse(TextResponse):

    @cached_property
    def bs(self):
        return BeautifulSoup(self.text, 'lxml')



import logging.config
from easy_spider.tool import EXE_PATH
from os import path


LOG_SETTING = {
    "version": 1,
    "formatters": {
        "default": {
            "format": "%(asctime)s %(levelname)-8s %(filename)s %(funcName)s %(lineno)d %(message)s",
            "datefmt": '%Y-%m-%d %H:%M:%S'
        },
        "colored": {
            '()': 'colorlog.ColoredFormatter',
            'format': "%(log_color)s[%(asctime)s] %(levelname)-8s %(filename)s %(funcName)s(%(lineno)d) %(message)s",
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "colored",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "filename": path.join(EXE_PATH, "easy_spider.log"),
            "level": "WARNING",
            "formatter": "default",
            "maxBytes": 1024,
            "encoding": "utf-8"
        }
    },
    "loggers": {
        "custom": {
            "level": "INFO",
            "handlers": ["console", "file"]
        }
    }
}
logging.config.dictConfig(LOG_SETTING)
logger = logging.getLogger("custom")



import posixpath
from typing import Any
from os import path
import sys


def get_extension(url):
    return posixpath.splitext(url)[-1].lower().lstrip(".")


def get_public_attr(instance: Any):
    for attr, value in instance.__dict__.items():
        if not attr.startswith("_"):
            yield attr


def copy_attr(attr, _from, _to):
    setattr(_to, attr, getattr(_from, attr))


def get_abs_path(file_path):
    return path.dirname(path.abspath(file_path))


ABS_PATH = get_abs_path(__file__)
sys.path.append(path.join(ABS_PATH, ".."))

EXE_PATH = path.dirname(path.abspath(__name__))



import unittest
from easy_spider import tool


class A:
    def __init__(self):
        self.a = 1
        self._a = 1


class B:
    def __init__(self):
        self.a = 2


class TestTool(unittest.TestCase):

    def test_extension(self):
        ext = tool.get_extension("http://localhost:5000/%E4%B8%AD%E5%9B%BD%E7%BA%A2.txt")
        self.assertEqual(ext, "txt")

    def test_get_public_attr(self):
        a = A()
        attrs = list(tool.get_public_attr(a))
        self.assertEqual(attrs, ["a"])

    def test_copy_public_attr(self):
        a = A()
        b = A()
        tool.copy_attr("a", b, a)
        self.assertEqual(a.a, b.a)


if __name__ == '__main__':
    unittest.main()



from easy_spider.network.client import AsyncClient
from easy_spider.core.task import AsyncTask
from easy_spider.core.spider import AsyncSpider
from aiohttp import ClientSession
from asyncio import get_event_loop
from cached_property import cached_property


class AsyncSpiderEvn:
    def __init__(self, loop=None):
        self._client = None
        self._loop = loop
        self._session = None

    def _prepare(self):
        if self._session is None:
            self._session = ClientSession()
        if self._client is None:
            self._client = AsyncClient()

    async def _run_spider(self, spider):
        self._prepare()
        spider.set_session(self._session)
        task = AsyncTask(spider)
        await task.run()

    @cached_property
    def loop(self):
        return self._loop or get_event_loop()

    def run(self, spider: AsyncSpider):
        self.loop.run_until_complete(self._run_spider(spider))

    def clear(self):
        self.__del__()

    def __del__(self):
        if not self._loop or self.loop.is_closed():
            return
        self._session and self.loop.run_until_complete(self._session.close())


async_env = AsyncSpiderEvn()
